package name.nathanmcrae.numbersstation;

import com.leakyabstractions.result.api.Result;
import com.leakyabstractions.result.core.Results;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.Optional;
import java.util.Random;
import java.util.ResourceBundle;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.fxml.Initializable;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.Label;
import javafx.scene.control.TextArea;
import javafx.scene.control.TextField;
import javafx.scene.image.Image;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.KeyCode;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.control.SplitPane;
import javafx.stage.Modality;
import javafx.stage.Stage;

public class MainController implements Initializable {
    private static final Logger logger = Logger.getLogger(Main.class.getName());

    private Stage aboutStage;
    private Stage helpStage;
    private Stage settingsStage;
    private Stage selectStationStage;
    private MainSettings settings;
    private StationSettings selectedStation;

    private StringProperty nextSendTime = new SimpleStringProperty();

    private StringProperty selectedStationName = new SimpleStringProperty();

    private SimpleBooleanProperty unsavedChanges = new SimpleBooleanProperty();

    public StringProperty windowTitle = new SimpleStringProperty("Numbers Station");

    @FXML
    private Label nextSendTimeLabel;

    @FXML
    private TextArea messageTextArea;

    @FXML
    private TextField stationNameField;

    @FXML
    private SplitPane splitPane;

    @FXML
    private void handleStationSettingsButtonPress() {
        if (settingsStage == null || !settingsStage.isShowing()) {
            try {
                FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource("/StationSettingsView.fxml"));
                Parent root = fxmlLoader.load();

                StationSettingsController controller = fxmlLoader.getController();
                if (selectedStation == null) {
                    Alert alert = new Alert(Alert.AlertType.WARNING);
                    alert.setTitle("Invalid Station Selected");
                    alert.setHeaderText(null);
                    alert.setContentText("Invalid station selected. Please choose a station.");
                    alert.showAndWait();
                    return;
                }
                controller.setStationSettings(selectedStation);
                controller.setStationNameList(settings.getStations().stream().map(StationSettings::toString).toList());

                settingsStage = new Stage();
                settingsStage.setMinWidth(600);
                settingsStage.setMinHeight(300);
                settingsStage.initModality(Modality.APPLICATION_MODAL);
                settingsStage.setTitle("Numbers Station Settings");
                settingsStage.getIcons().add(new Image(getClass().getResourceAsStream("/icon.png")));
                settingsStage.setScene(new Scene(root));
                settingsStage.show();
                settingsStage.setOnHiding(event -> {
                    updateStationSettings(selectedStation);
                });
            } catch (Exception e) {
                logger.log(Level.SEVERE, "Exception while opening settings", e);
            }
        } else {
            settingsStage.toFront();
        }
    }

    @FXML
    private void handleSelectStationButtonPress() throws Exception {
        if (selectStationStage == null || !selectStationStage.isShowing()) {
            try {
                FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource("/StationSelectionView.fxml"));
                Parent root = fxmlLoader.load();

                StationSelectionController controller = fxmlLoader.getController();
                controller.setSettings(settings);

                selectStationStage = new Stage();
                selectStationStage.setMinWidth(300);
                selectStationStage.setMinHeight(200);
                selectStationStage.initModality(Modality.APPLICATION_MODAL);
                selectStationStage.setUserData(null);
                selectStationStage.setTitle("Numbers Station Selection");
                selectStationStage.getIcons().add(new Image(getClass().getResourceAsStream("/icon.png")));

                Scene scene = new Scene(root);
                scene.addEventFilter(KeyEvent.KEY_PRESSED, (e) -> controller.handleKeyPressed(e));
                selectStationStage.setScene(scene);
                selectStationStage.show();
                selectStationStage.setOnHiding(event -> {
                    String newStationName = (String)selectStationStage.getUserData();
                    if (newStationName == null) {
                        return;
                    }
                    if (unsavedChanges.get()) {
                        if (!confirmSaveChanges()) {
                            return;
                        }
                    }
                    StationSettings newSelectedStation = settings.getStations().stream()
                            .filter(station -> station.getName().equals(newStationName))
                            .findFirst()
                            .orElse(null);
                    updateStationSettings(newSelectedStation);
                });
            } catch (Exception e) {
                logger.log(Level.SEVERE, "Exception while selecting station", e);
            }
        } else {
            selectStationStage.toFront();
        }
    }

    @FXML
    private void handleSetAsNextMessageButtonPress() {
        saveMessage();
    }

    private void saveMessage() {
        try {
            if (!Files.exists(selectedStation.stationPath())) {
                Files.createDirectories(selectedStation.stationPath());
            }

            Files.write(selectedStation.nextMessagePath(), messageTextArea.getText().getBytes());
            unsavedChanges.set(false);
        } catch (IOException e) {
            String message = "Failed to write next message to " + selectedStation.nextMessagePath().toString();
            logger.log(Level.SEVERE, message, e);
            Alert alert = new Alert(Alert.AlertType.ERROR);
            alert.setTitle("Exception");
            alert.setHeaderText(null);
            alert.setContentText(message);
            alert.showAndWait();
        }
    }

    @Override
    public void initialize(URL location, ResourceBundle resources) {
        stationNameField.textProperty().bindBidirectional(selectedStationName);
        nextSendTimeLabel.textProperty().bind(nextSendTime);

        Result<MainSettings, Exception> result = MainSettings.load();
        if (!result.hasSuccess()) {
            String logPath = Main.getStatePath().resolve("main.log").toString();
            String message = "Unable to load settings file. See log file at " + logPath + " for details. Reinitialize settings? This may overwrite existing settings.";
            Alert alert = new Alert(Alert.AlertType.ERROR, message, ButtonType.NO, ButtonType.YES);
            alert.setTitle("Settings load error");
            alert.setHeaderText(null);

            //Deactivate Defaultbehavior for yes-Button:
            Button yesButton = (Button) alert.getDialogPane().lookupButton( ButtonType.YES );
            yesButton.setDefaultButton( false );

            //Activate Defaultbehavior for no-Button:
            Button noButton = (Button) alert.getDialogPane().lookupButton( ButtonType.NO );
            noButton.setDefaultButton( true );

            Optional<ButtonType> promptResult = alert.showAndWait();

            if (!promptResult.isPresent()) {
                System.exit(1);
            }

            if (promptResult.get() == ButtonType.NO) {
                System.exit(1);
            }

            settings = new MainSettings();
        } else {
            settings = result.getSuccess().get();
        }

        selectedStationName.set(settings.getSelectedStationName());

        if (selectedStationName.get() == null || selectedStationName.get() == "") {
            selectedStation = settings.getStations().stream().findFirst().orElse(null);
            selectedStationName.set(selectedStation.getName());
        }

        selectedStation = settings.getStations().stream()
            .filter(station -> station.getName().equals(selectedStationName.get()))
            .findFirst()
            .orElse(null);

        if (selectedStation == null) {
            logger.log(Level.SEVERE, "Selected station '" + selectedStationName.get() + "' not found");
            selectedStation = settings.getStations().get(0);
            selectedStationName.set(selectedStation.getName());
        }

        updateStationSettings(selectedStation);

        messageTextArea.addEventFilter(KeyEvent.ANY, event -> {
            int cursorPosition = messageTextArea.getCaretPosition();

            String character = event.getCharacter();
            KeyCode code = event.getCode();

            String str = event.getText();

            // Consume the event to block the default behavior
            if (!(code == KeyCode.LEFT ||
                  code == KeyCode.RIGHT ||
                  code == KeyCode.UP ||
                  code == KeyCode.DOWN ||
                  code == KeyCode.HOME ||
                  code == KeyCode.END ||
                  code == KeyCode.PAGE_UP ||
                  code == KeyCode.PAGE_DOWN)) {
                event.consume();
            }

            if (event.getEventType() == KeyEvent.KEY_PRESSED &&
                (code == KeyCode.DIGIT0 ||
                 code == KeyCode.DIGIT1 ||
                 code == KeyCode.DIGIT2 ||
                 code == KeyCode.DIGIT3 ||
                 code == KeyCode.DIGIT4 ||
                 code == KeyCode.DIGIT5 ||
                 code == KeyCode.DIGIT6 ||
                 code == KeyCode.DIGIT7 ||
                 code == KeyCode.DIGIT8 ||
                 code == KeyCode.DIGIT9)) {
                unsavedChanges.set(true);

                if (cursorPosition < messageTextArea.getLength()) {
                    char currentChar = messageTextArea.getText().charAt(cursorPosition);
                    if (Character.isWhitespace(currentChar)) {
                        messageTextArea.replaceText(cursorPosition + 1, cursorPosition + 2, str);
                    } else {
                        messageTextArea.replaceText(cursorPosition, cursorPosition + 1, str);
                    }
                }
            }
        });

        // The SplitPane allows the user to set the number of columns of
        // numbers, which can help align with their pads.
        //
        // The splitpane doesn't seem to stick at 1.0 when resizing though,
        // so we snap it to the right when it's close.
        splitPane.widthProperty().addListener((obs, oldVal, newVal) -> {
            if (splitPane.getDividerPositions()[0] > 0.95) {
                splitPane.setDividerPositions(1.0);
            }
        });

        unsavedChanges.addListener((obs2, wasUnsaved, isNowUnsaved) -> {
            if (isNowUnsaved) {
                if (!windowTitle.get().startsWith("*")) {
                    windowTitle.set("*" + windowTitle.get());
                }
            } else {
                if (windowTitle.get().startsWith("*")) {
                    windowTitle.set(windowTitle.get().substring(1, windowTitle.get().length()));
                }
            }
        });
    }

    private void updateStationSettings(StationSettings newStationSettings) {
        if (newStationSettings == null) {
            return;
        }

        logger.info("Updating station settings: " + newStationSettings.getName());

        DateTimeFormatter dateFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ss");

        try {
        nextSendTime.set(newStationSettings.nextSendTime().format(dateFormatter));
        } catch (StationSettings.StationSettingsException e) {
            String errMessage = "Failed to calculate next message time";
            logger.log(Level.SEVERE, errMessage, e);
            Alert alert = new Alert(Alert.AlertType.ERROR);
            alert.setTitle("Exception");
            alert.setHeaderText(null);
            alert.setContentText(errMessage + ": " + e.getMessage());
            alert.showAndWait();
            return;
        }

        selectedStation = newStationSettings;
        selectedStationName.set(newStationSettings.getName());
        settings.setSelectedStationName(selectedStationName.get());

        settings.save();

        Path nextMessagePath = selectedStation.stationPath().resolve("next-message.txt");

        String messageText;
        if (!Files.exists(nextMessagePath)) {
            try {
                messageText = selectedStation.generateMessage(settings.getMessageGenerationAttempts());
            } catch (StationSettings.MessageGenerationException e) {
                String errMessage = "Failed to generate message";
                logger.log(Level.SEVERE, errMessage, e);
                Alert alert = new Alert(Alert.AlertType.ERROR);
                alert.setTitle("Exception");
                alert.setHeaderText(null);
                alert.setContentText(errMessage + ": " + e.getMessage());
                alert.showAndWait();
                return;
            }

            try {
                if (!Files.exists(selectedStation.stationPath())) {
                    Files.createDirectories(selectedStation.stationPath());
                }
                Files.write(nextMessagePath, messageText.getBytes(StandardCharsets.UTF_8));
            } catch (IOException e) {
                String errMessage = "Failed to save message";
                logger.log(Level.SEVERE, errMessage, e);
                Alert alert = new Alert(Alert.AlertType.ERROR);
                alert.setTitle("Exception");
                alert.setHeaderText(null);
                alert.setContentText(errMessage + ": " + e.getMessage());
                alert.showAndWait();
                return;
            }
        } else {
            try {
                messageText = new String(Files.readAllBytes(nextMessagePath));
            } catch (IOException e) {
                String errMessage = "Failed to read message";
                logger.log(Level.SEVERE, errMessage, e);
                Alert alert = new Alert(Alert.AlertType.ERROR);
                alert.setTitle("Exception");
                alert.setHeaderText(null);
                alert.setContentText(errMessage + ": " + e.getMessage());
                alert.showAndWait();
                return;
            }
        }

        unsavedChanges.set(false);
        messageTextArea.setText(messageText);
    }

    /**
     * @return Whether to continue with the action that prompted the confirmation
     */
    public boolean confirmSaveChanges() {
        Alert alert = new Alert(Alert.AlertType.CONFIRMATION, "Do you want to save your changes to the message?", ButtonType.YES, ButtonType.NO, ButtonType.CANCEL);
        alert.setHeaderText(null);
        Optional<ButtonType> result = alert.showAndWait();

        if (!result.isPresent()) {
            return false;
        }

        if (result.get() == ButtonType.YES) {
            saveMessage();
            return true;
        } else if (result.get() == ButtonType.NO) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * @return whether close should continue or not
     */
    public boolean handleCloseRequest() {
        if (unsavedChanges.get()) {
            return confirmSaveChanges();
        }

        return true;
    }

    public void handleAboutButtonPress() {
        try {
            if (aboutStage == null) {
            FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource("/AboutView.fxml"));
                Parent root = fxmlLoader.load();
                aboutStage = new Stage();

                aboutStage.setMinWidth(250);
                aboutStage.setMinHeight(230);

                aboutStage.setTitle("About Numbers Station");
                aboutStage.getIcons().add(new Image(getClass().getResourceAsStream("/icon.png")));
                aboutStage.setScene(new Scene(root));
            }

            if (aboutStage.isShowing()) {
                aboutStage.toFront();
            } else {
                aboutStage.show();
            }
        } catch (IOException e) {
            logger.log(Level.SEVERE, "Exception while opening application about screen", e);
        }
    }

    public void handleHelpButtonPress() {
        try {
            if (helpStage == null) {
            FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource("/HelpView.fxml"));
                Parent root = fxmlLoader.load();
                helpStage = new Stage();

                helpStage.setMinWidth(400);
                helpStage.setMinHeight(400);

                helpStage.setTitle("Numbers Station Help");
                helpStage.getIcons().add(new Image(getClass().getResourceAsStream("/icon.png")));
                helpStage.setScene(new Scene(root));
            }

            if (helpStage.isShowing()) {
                helpStage.toFront();
            } else {
                helpStage.show();
            }
        } catch (IOException e) {
            logger.log(Level.SEVERE, "Exception while opening help", e);
        }
    }
}
