package name.nathanmcrae.numbersstation;

import java.io.IOException;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.Optional;
import javafx.collections.ListChangeListener;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.control.Alert;
import javafx.scene.control.ButtonType;
import javafx.scene.control.ListCell;
import javafx.scene.control.ListView;
import javafx.scene.control.TextInputDialog;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.event.ActionEvent;
import javafx.event.Event;
import javafx.scene.Node;
import javafx.stage.Stage;

public class StationSelectionController {
    private static final Logger logger = Logger.getLogger(Main.class.getName());

    private MainSettings settings;

    @FXML
    private ListView<StationSettings> stationListView;

    private ObservableList<StationSettings> stationList;

    public StationSelectionController() {
        stationList = FXCollections.observableArrayList();
        stationList.addListener((ListChangeListener<StationSettings>) change -> onStationListChanged(change));
    }

    @FXML
    private void initialize() {
        // Set custom cell factory to display station name
        stationListView.setCellFactory(param -> new ListCell<StationSettings>() {
            @Override
            protected void updateItem(StationSettings item, boolean empty) {
                super.updateItem(item, empty);
                if (empty || item == null || item.getName() == null) {
                    setText(null);
                } else {
                    setText(item.getName());
                }
            }
        });
    }

    public void setSettings(MainSettings newSettings) {
        settings = newSettings;
        stationList.addAll(settings.getStations());
        stationListView.setItems(stationList);
    }

    @FXML
    private void handleAddButtonPress(ActionEvent event) {
        TextInputDialog dialog = new TextInputDialog();
        dialog.setTitle("New Station");
        dialog.setHeaderText("Add a New Station");
        dialog.setContentText("Please enter the station name:");

        Optional<String> result = dialog.showAndWait();
        result.ifPresent(stationName -> {
            if (stationName == null || stationName.equals("")) {
                Alert alert = new Alert(Alert.AlertType.WARNING);
                alert.setTitle("Invalid station name");
                alert.setHeaderText(null);
                alert.setContentText("Station name must not be empty");
                alert.showAndWait();
                return;
            }

            for (StationSettings station : stationList) {
                if (stationName.toLowerCase().equals(station.getName().toLowerCase())) {
                    Alert alert = new Alert(Alert.AlertType.WARNING);
                    alert.setTitle("Invalid station name");
                    alert.setHeaderText(null);
                    alert.setContentText("Station name '" + stationName + "' is already used.");
                    alert.showAndWait();
                    return;
                }
            }

            StationSettings newStation = new StationSettings(stationName);
            stationList.add(newStation);
        });
    }

    @FXML
    private void handleCancelButtonPress(Event event) {
        Node node = (Node) event.getSource();
        Stage stage = (Stage) node.getScene().getWindow();
        stage.close();
    }

    @FXML
    public void handleKeyPressed(KeyEvent event) {
        KeyCode code = event.getCode();
        if (code == KeyCode.ENTER) {
            selectHighlightedStation();
        } else if (code == KeyCode.ESCAPE) {
            Stage stage = (Stage) stationListView.getScene().getWindow();
            stage.close();
        }
    }

    @FXML
    private void handleRemoveButtonPress(ActionEvent event) {
        int selectedIndex = stationListView.getSelectionModel().getSelectedIndex();
        if (selectedIndex < 0) {
            return;
        }
        StationSettings station = stationList.get(selectedIndex);
        String stationName = station.toString();

        String message = "Delete station '" + stationName + "'? All data will be erased.";
        Alert alert = new Alert(Alert.AlertType.CONFIRMATION, message, ButtonType.YES, ButtonType.NO);
        alert.setTitle("Delete station?");
        alert.setHeaderText(null);
        alert.showAndWait();

        if (selectedIndex >= 0 && alert.getResult() == ButtonType.YES) {
            stationList.remove(selectedIndex);
        }
    }

    @FXML
    private void handleSelectButtonPress(Event e) {
        selectHighlightedStation();
    }

    private void onStationListChanged(ListChangeListener.Change<? extends StationSettings> change) {
        while (change.next()) {
            if (change.wasAdded()) {
                for (StationSettings addedStation : change.getAddedSubList()) {
                }
            }
            if (change.wasRemoved()) {
                for (StationSettings removedStation : change.getRemoved()) {
                    try {
                        StationSettings.deleteStationData(removedStation.getName());
                        removedStation.deleteDir();
                    } catch (IOException e) {
                        logger.log(Level.SEVERE, "Exception when removing station directory", e);
                        Alert alert = new Alert(Alert.AlertType.ERROR);
                        alert.setTitle("Error removing station");
                        alert.setHeaderText(null);
                        alert.setContentText("Exception when removing station directory: " + e.getMessage());
                        alert.showAndWait();
                        return;
                    }
                }
            }
        }
        settings.getStations().clear();
        settings.getStations().addAll(stationList);
        settings.save();
    }

    private void selectHighlightedStation() {
        Stage stage = (Stage) stationListView.getScene().getWindow();
        StationSettings selectedStation = stationListView.getSelectionModel().getSelectedItem();

        if (selectedStation != null) {
            stage.setUserData(selectedStation.getName());
        }
        stage.close();
    }
}
