package name.nathanmcrae.numbersstation;

import com.leakyabstractions.result.api.Result;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.Files;
import java.nio.file.FileVisitResult;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.SimpleFileVisitor;
import java.security.SecureRandom;
import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZonedDateTime;
import java.time.ZoneId;
import java.time.LocalTime;
import java.time.Period;
import java.util.ArrayList;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.Random;

public class StationSettings {
    private static final Logger logger = Logger.getLogger(Main.class.getName());

    private String address;
    private int digitsPerGroup;
    private String externalProgramCommand;
    private Duration externalProgramTimeout;
    private boolean manageScheduleExternally;
    private int messageLength;
    private MessageMethod messageMethod;
    private MessagePeriod messagePeriod;
    private String name;
    private String password;
    private ArrayList<String> prefixes;
    private ZonedDateTime scheduleStart;
    private String username;

    public enum MessageMethod {
        EXTERNAL_PROGRAM,
        SFTP,
        WORDPRESS
    }

    public enum MessagePeriod {
        DAILY,
        WEEKLY,
        MONTHLY
    }

    public StationSettings() {
        prefixes = new ArrayList<String>();
        digitsPerGroup = 4;
        messageLength = 100;
        externalProgramTimeout = Duration.ofSeconds(5);
        messageMethod = MessageMethod.SFTP;
        messagePeriod = MessagePeriod.DAILY;
        scheduleStart = ZonedDateTime.now();
    }

    public StationSettings(String newName) {
        name = newName;
        prefixes = new ArrayList<String>();
        digitsPerGroup = 4;
        externalProgramTimeout = Duration.ofSeconds(5);
        messageLength = 100;
        messageMethod = MessageMethod.SFTP;
        messagePeriod = MessagePeriod.DAILY;
        scheduleStart = ZonedDateTime.now();
    }

    public static void deleteStationData(String stationName) throws IOException {
        StationSettings temp = new StationSettings(stationName);

        temp.deleteDir();

        String osName = System.getProperty("os.name").toLowerCase();
        if (osName.contains("win")) {
            WindowsScheduler.removeSchedule(temp);
        } else if (osName.contains("nix") || osName.contains("nux") || osName.contains("aix")) {
            LinuxScheduler.removeSchedule(temp);
        } else {
            logger.log(Level.SEVERE, "Unsupported OS " + osName);
        }
    }

    public static boolean stationDataExists(String stationName) {
        StationSettings temp = new StationSettings(stationName);
        if (Files.exists(temp.stationPath())) {
            return true;
        }

        String osName = System.getProperty("os.name").toLowerCase();
        if (osName.contains("win")) {
            Result<Boolean, String> result = WindowsScheduler.scheduleExists(temp);
            if (result.hasSuccess() && result.getSuccess().get()) {
                return true;
            }
        } else if (osName.contains("nix") || osName.contains("nux") || osName.contains("aix")) {
            Result<Boolean, String> result = LinuxScheduler.scheduleExists(temp);
            if (result.hasSuccess() && result.getSuccess().get()) {
                return true;
            }
        } else {
            logger.log(Level.SEVERE, "Unsupported OS " + osName);
        }

        return false;
    }

    public void deleteDir() throws IOException {
        if (!Files.exists(stationPath())) {
            return;
        }

        Files.walkFileTree(stationPath(), new SimpleFileVisitor<Path>() {
            @Override
            public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                Files.delete(file);
                return FileVisitResult.CONTINUE;
            }

            @Override
            public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
                Files.delete(dir);
                return FileVisitResult.CONTINUE;
            }
        });
    }

    public Path nextMessagePath() {
        return stationPath().resolve("next-message.txt");
    }

    public Path stationPath() {
        String stationDirName = null;
        try {
            stationDirName = java.net.URLEncoder.encode(name, "UTF-8");
        } catch(UnsupportedEncodingException e) {} // We know UTF-8 is supported

        return Main.getConfigPath().resolve(stationDirName);
    }

    public String generateMessage(int generationAttempts) throws MessageGenerationException {
        SecureRandom random = new SecureRandom();
        StringBuilder messageBuilder = new StringBuilder();

        int attemptsLeft = generationAttempts;
        boolean noPrefixCollisions = false;

        while (!noPrefixCollisions) {
            messageBuilder = new StringBuilder();

            for (int i = 0; i < messageLength; i++) {
                messageBuilder.append(random.nextInt(10));
            }

            String message = messageBuilder.toString();

            boolean prefixCollision = false;

            for (String prefix : prefixes) {
                if (message.startsWith(prefix)) {
                    prefixCollision = true;
                }
            }

            if (!prefixCollision) {
                noPrefixCollisions = true;
            }

            attemptsLeft--;
            if (prefixCollision && attemptsLeft <= 0) {
                throw new MessageGenerationException("Failed " + generationAttempts + " attempts to generate message without matching prefixes. Ensure prefixes are valid.");
            }
        }

        String messageNoSpaces = messageBuilder.toString();
        StringBuilder messageFormattedBuilder = new StringBuilder();

        for (int i = 0; i < messageNoSpaces.length(); i++) {
            if (i > 0 && i % digitsPerGroup == 0) {
                messageFormattedBuilder.append(" ");
            }
            messageFormattedBuilder.append(messageNoSpaces.charAt(i));
        }

        return messageFormattedBuilder.toString();
    }

    public LocalDateTime nextSendTime() throws StationSettingsException {
        LocalDate startDate = scheduleStart.withZoneSameInstant(ZoneId.systemDefault()).toLocalDate();
        LocalTime startTime = scheduleStart.withZoneSameInstant(ZoneId.systemDefault()).toLocalTime();
        Period sinceScheduleStart = Period.between(startDate, LocalDate.now());
        // If this period's time has not yet passed, then show that time.
        // Otherwise, show the next period's time.
        switch (messagePeriod) {
            case DAILY:
                if (LocalTime.now().isBefore(startTime)) {
                    return LocalDate.now().atTime(startTime);
                } else {
                    return LocalDate.now().plusDays(1).atTime(startTime);
                }
            case WEEKLY:
                int weekdayDifference = startDate.getDayOfWeek().getValue() - LocalDate.now().getDayOfWeek().getValue();
                if (weekdayDifference > 0) {
                    return LocalDate.now().plusDays(weekdayDifference).atTime(startTime);
                 } else if (weekdayDifference == 0){
                    if (LocalTime.now().isBefore(startTime)) {
                        return LocalDate.now().atTime(startTime);
                    } else {
                        return LocalDate.now().plusWeeks(1).atTime(startTime);
                    }
                } else {
                    return LocalDate.now().plusDays(7 + weekdayDifference).atTime(startTime);
                }
            case MONTHLY:
                int monthdayDifference = startDate.getDayOfMonth() - LocalDate.now().getDayOfMonth();
                if (monthdayDifference > 0) {
                    return LocalDate.now().plusDays(monthdayDifference).atTime(startTime);
                } else if (monthdayDifference == 0) {
                    if (LocalTime.now().isBefore(startTime)) {
                        return LocalDate.now().atTime(startTime);
                    } else {
                        return LocalDate.now().plusMonths(1).atTime(startTime);
                    }
                } else {
                    return LocalDate.now().plusMonths(1).plusDays(monthdayDifference).atTime(startTime);
                }
            default:
                throw new StationSettingsException("Invalid period value: " + messagePeriod);
        }
    }

    public String getAddress() {
        return address;
    }

    public void setAddress(String newAddress) {
        address = newAddress;
    }

    public int getDigitsPerGroup() {
        return digitsPerGroup;
    }

    public void setDigitsPerGroup(int newDigitsPerGroup) {
        digitsPerGroup = newDigitsPerGroup;
    }

    public String getExternalProgramCommand() {
        return externalProgramCommand;
    }

    public void setExternalProgramCommand(String newExternalProgramCommand) {
        externalProgramCommand = newExternalProgramCommand;
    }

    public Duration getExternalProgramTimeout() {
        return externalProgramTimeout;
    }

    public void setExternalProgramTimeout(Duration newExternalProgramTimeout) {
        externalProgramTimeout = newExternalProgramTimeout;
    }

    public boolean getManageScheduleExternally() {
        return manageScheduleExternally;
    }

    public void setManageScheduleExternally(boolean newManageScheduleExternally) {
        manageScheduleExternally = newManageScheduleExternally;
    }

    public int getMessageLength() {
        return messageLength;
    }

    public void setMessageLength(int newMessageLength) {
        messageLength = newMessageLength;
    }

    public MessageMethod getMessageMethod() {
        return messageMethod;
    }

    public void setMessageMethod(MessageMethod newMessageMethod) {
        messageMethod = newMessageMethod;
    }

    public MessagePeriod getMessagePeriod() {
        return messagePeriod;
    }

    public void setMessagePeriod(MessagePeriod newMessagePeriod) {
        messagePeriod = newMessagePeriod;
    }

    public String getName() {
        return name;
    }

    public void setName(String newName) {
        name = newName;
    }

    public String getPassword() {
        return password;
    }

    public void setPassword(String newPassword) {
        password = newPassword;
    }

    public ArrayList<String> getPrefixes() {
        if (prefixes == null) {
            prefixes = new ArrayList<String>();
        }
        return prefixes;
    }

    public ZonedDateTime getScheduleStart() {
        return scheduleStart;
    }

    public void setScheduleStart(ZonedDateTime newScheduleStart) {
        scheduleStart = newScheduleStart;
    }

    public String getUsername() {
        return username;
    }

    public void setUsername(String newUsername) {
        username = newUsername;
    }

    @Override
    public String toString() {
        return name;
    }

    public class MessageGenerationException extends Exception {
        public MessageGenerationException(String message) {
            super(message);
        }
    }

    public static class StationSettingsException extends Exception {
        public StationSettingsException(String message) {
            super(message);
        }
    }

}
