package name.nathanmcrae.numbersstation;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.dataformat.xml.XmlMapper;
import com.jcraft.jsch.JSch;
import com.jcraft.jsch.JSchException;
import com.jcraft.jsch.Session;
import com.tearsofaunicorn.wordpress.api.model.Post;
import com.tearsofaunicorn.wordpress.api.WordpressClient;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.format.DateTimeFormatter;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.ZonedDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.Optional;
import javafx.beans.binding.Bindings;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.IntegerProperty;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleIntegerProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.Event;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.control.Alert;
import javafx.scene.control.Button;
import javafx.scene.control.CheckBox;
import javafx.scene.control.DatePicker;
import javafx.scene.control.Label;
import javafx.scene.control.ListCell;
import javafx.scene.control.ListView;
import javafx.scene.control.RadioButton;
import javafx.scene.control.Spinner;
import javafx.scene.control.SpinnerValueFactory;
import javafx.scene.control.TextField;
import javafx.scene.control.TextInputDialog;
import javafx.scene.control.ToggleGroup;
import javafx.scene.image.Image;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.KeyCode;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.stage.Modality;
import javafx.stage.Stage;
import javafx.util.Callback;

public class StationSettingsController {
    public enum ConnectionStatus {
            SUCCESS, FAILURE, NEUTRAL
    }

    private static final Logger logger = Logger.getLogger(Main.class.getName());

    private ObjectProperty<ConnectionStatus> connectionTestStatus = new SimpleObjectProperty<>(ConnectionStatus.NEUTRAL);
    private StringProperty connectionTestDescription = new SimpleStringProperty();
    private IntegerProperty digitsPerGroup = new SimpleIntegerProperty();
    private StringProperty externalProgramCommand = new SimpleStringProperty();
    private BooleanProperty manageScheduleExternally = new SimpleBooleanProperty();
    private IntegerProperty messageLength = new SimpleIntegerProperty();
    private ObjectProperty<StationSettings.MessageMethod> messageMethod = new SimpleObjectProperty<>();
    private ObjectProperty<StationSettings.MessagePeriod> messagePeriod = new SimpleObjectProperty<>();
    private StringProperty password = new SimpleStringProperty();
    private StationSettings settings;
    private StringProperty stationAddress = new SimpleStringProperty();
    private StringProperty stationName = new SimpleStringProperty();
    private ObjectProperty<LocalDate> scheduleStartDate = new SimpleObjectProperty<>();
    private StringProperty scheduleStartTime = new SimpleStringProperty();
    private StringProperty username = new SimpleStringProperty();

    private ArrayList<String> stationNames = new ArrayList<String>();

    @FXML
    private Label connectionTestStatusLabel;

    @FXML
    private TextField stationNameField;

    @FXML
    private TextField stationAddressField;

    @FXML
    private RadioButton dailyRadioButton;

    @FXML
    private Spinner<Integer> digitsPerGroupSpinner;

    @FXML
    private TextField externalProgramCommandField;

    @FXML
    private CheckBox manageScheduleExternallyCheckBox;

    @FXML
    private Spinner<Integer> messageLengthSpinner;

    @FXML
    private ToggleGroup messageMethodGroup;

    @FXML
    private ToggleGroup messagePeriodGroup;

    @FXML
    private RadioButton monthlyRadioButton;

    @FXML
    private RadioButton sftpRadioButton;

    @FXML
    private TextField scheduleStartTimeField;

    @FXML
    private Button testConnectionButton;

    @FXML
    private RadioButton externalProgramRadioButton;

    @FXML
    private TextField usernameField;

    @FXML
    private TextField passwordField;

    @FXML
    private ListView<String> prefixListView;

    @FXML
    private DatePicker scheduleStartDatePicker;

    @FXML
    private RadioButton weeklyRadioButton;

    @FXML
    private RadioButton wordpressRadioButton;

    public StationSettingsController() { }

    @FXML
    private void initialize() {
        connectionTestStatusLabel.styleProperty().bind(Bindings.createStringBinding(() -> {
            switch (connectionTestStatus.get()) {
                case SUCCESS:
                    return "-fx-text-fill: green;";
                case FAILURE:
                    return "-fx-text-fill: red;";
                default:
                    return "-fx-text-fill: black;";
            }
        }, connectionTestStatus));
        connectionTestStatusLabel.textProperty().bindBidirectional(connectionTestDescription);
        externalProgramCommandField.textProperty().bindBidirectional(externalProgramCommand);
        manageScheduleExternallyCheckBox.selectedProperty().bindBidirectional(manageScheduleExternally);
        stationNameField.textProperty().bindBidirectional(stationName);
        stationAddressField.textProperty().bindBidirectional(stationAddress);
        passwordField.textProperty().bindBidirectional(password);
        scheduleStartDatePicker.valueProperty().bindBidirectional(scheduleStartDate);
        scheduleStartTimeField.textProperty().bindBidirectional(scheduleStartTime);
        usernameField.textProperty().bindBidirectional(username);

        digitsPerGroupSpinner.setValueFactory(new SpinnerValueFactory.IntegerSpinnerValueFactory(1, 100, 1));
        messageLengthSpinner.setValueFactory(new SpinnerValueFactory.IntegerSpinnerValueFactory(1, 1000, 1));

        digitsPerGroupSpinner.getValueFactory().valueProperty().bindBidirectional(digitsPerGroup.asObject());
        messageLengthSpinner.getValueFactory().valueProperty().bindBidirectional(messageLength.asObject());

        scheduleStartTimeField.addEventFilter(KeyEvent.ANY, event -> {
            int cursorPosition = scheduleStartTimeField.getCaretPosition();

            String character = event.getCharacter();
            KeyCode code = event.getCode();

            String str = event.getText();

            // Consume the event to block the default behavior
            if (!(code == KeyCode.LEFT ||
                  code == KeyCode.RIGHT ||
                  code == KeyCode.HOME ||
                  code == KeyCode.END)) {
                event.consume();
            }

            if (event.getEventType() == KeyEvent.KEY_PRESSED &&
                (code == KeyCode.UP || code == KeyCode.DOWN)) {
                boolean up = false;
                if (code == KeyCode.UP) {
                    up = true;
                }

                String[] timeParts = scheduleStartTimeField.getText().split(":");
                if (timeParts.length == 3) {
                    int hours = Integer.parseInt(timeParts[0]);
                    int minutes = Integer.parseInt(timeParts[1]);
                    int seconds = Integer.parseInt(timeParts[2]);

                    if (cursorPosition < 3) {
                        if (up && hours < 23) {
                            hours++;
                        } else if (!up && hours > 0) {
                            hours--;
                        }
                    } else if (cursorPosition < 6) {
                        if (up && minutes < 59) {
                            minutes++;
                        } else if (!up && minutes > 0) {
                            minutes--;
                        }
                    } else {
                        if (up && seconds < 59) {
                            seconds++;
                        } else if (!up && seconds > 0) {
                            seconds--;
                        }
                    }

                    scheduleStartTimeField.setText(String.format("%02d:%02d:%02d", hours, minutes, seconds));
                    scheduleStartTimeField.positionCaret(cursorPosition);
                }
            }

            if (event.getEventType() == KeyEvent.KEY_PRESSED &&
                (code == KeyCode.DIGIT0 ||
                 code == KeyCode.DIGIT1 ||
                 code == KeyCode.DIGIT2 ||
                 code == KeyCode.DIGIT3 ||
                 code == KeyCode.DIGIT4 ||
                 code == KeyCode.DIGIT5 ||
                 code == KeyCode.DIGIT6 ||
                 code == KeyCode.DIGIT7 ||
                 code == KeyCode.DIGIT8 ||
                 code == KeyCode.DIGIT9)) {

                 String editedTime = scheduleStartTimeField.getText();

                if (cursorPosition < scheduleStartTimeField.getLength()) {
                    char currentChar = scheduleStartTimeField.getText().charAt(cursorPosition);
                    if (currentChar == ':') {
                        cursorPosition++;
                    }
                    editedTime = editedTime.substring(0, cursorPosition) + str + editedTime.substring(cursorPosition + 1);
                }

                // Validate the time format
                String[] timeParts = editedTime.split(":");
                if (timeParts.length == 3) {
                    int hours = Integer.parseInt(timeParts[0]);
                    int minutes = Integer.parseInt(timeParts[1]);
                    int seconds = Integer.parseInt(timeParts[2]);

                    if (hours > 23) {
                        hours = 23;
                        editedTime = "23" + editedTime.substring(2);
                    }

                    if (!(hours < 0 || hours > 23 || minutes < 0 || minutes > 59 || seconds < 0 || seconds > 59)) {
                        scheduleStartTimeField.setText(editedTime);
                        scheduleStartTimeField.positionCaret(cursorPosition + 1);
                    }
                } else {
                    scheduleStartTimeField.setText("00:00:00");
                }
            }
        });

        messageMethodGroup.selectedToggleProperty().addListener((observable, oldValue, newValue) -> {
            if (newValue == externalProgramRadioButton) {
                messageMethod.set(StationSettings.MessageMethod.EXTERNAL_PROGRAM);
            } else if (newValue == sftpRadioButton) {
                messageMethod.set(StationSettings.MessageMethod.SFTP);
            } else if (newValue == wordpressRadioButton) {
                messageMethod.set(StationSettings.MessageMethod.WORDPRESS);
            }
        });

        messageMethod.addListener((observable, oldValue, newValue) -> {
            if (newValue == StationSettings.MessageMethod.EXTERNAL_PROGRAM) {
                messageMethodGroup.selectToggle(externalProgramRadioButton);
            } else if (newValue == StationSettings.MessageMethod.SFTP) {
                messageMethodGroup.selectToggle(sftpRadioButton);
            } else if (newValue == StationSettings.MessageMethod.WORDPRESS) {
                messageMethodGroup.selectToggle(wordpressRadioButton);
            }
        });

        messagePeriodGroup.selectedToggleProperty().addListener((observable, oldValue, newValue) -> {
            if (newValue == dailyRadioButton) {
                messagePeriod.set(StationSettings.MessagePeriod.DAILY);
            } else if (newValue == monthlyRadioButton) {
                messagePeriod.set(StationSettings.MessagePeriod.MONTHLY);
            } else if (newValue == weeklyRadioButton) {
                messagePeriod.set(StationSettings.MessagePeriod.WEEKLY);
            } 
        });

        messagePeriod.addListener((observable, oldValue, newValue) -> {
            if (newValue == StationSettings.MessagePeriod.DAILY) {
                messagePeriodGroup.selectToggle(dailyRadioButton);
            } else if (newValue == StationSettings.MessagePeriod.MONTHLY) {
                messagePeriodGroup.selectToggle(monthlyRadioButton);
            } else if (newValue == StationSettings.MessagePeriod.WEEKLY) {
                messagePeriodGroup.selectToggle(weeklyRadioButton);
            } 
        });

        // Set user data. Default result is no update.
        stationNameField.sceneProperty().addListener((observable, oldScene, newScene) -> {
            if (newScene != null) {
                Stage stage = (Stage) newScene.getWindow();
                stage.setUserData(false);
            }
        });

        prefixListView.setCellFactory(new Callback<ListView<String>, ListCell<String>>() {
            @Override
            public ListCell<String> call(ListView<String> listView) {
                return new ListCell<String>() {
                    @Override
                    protected void updateItem(String item, boolean empty) {
                        super.updateItem(item, empty);
                        if (empty || item == null) {
                            setText(null);
                        } else {
                            String newText = item.replaceAll(" ", "");
                            StringBuilder sb = new StringBuilder();

                            for (int i = 0; i < newText.length(); i++) {
                                sb.append(newText.charAt(i));
                                if ((i + 1) % (digitsPerGroup.get()) == 0) {
                                    sb.append(' ');
                                }
                            }
                            setText(sb.toString());
                        }
                    }
                };
            }
        });

        externalProgramCommandField.disableProperty().bind(externalProgramRadioButton.selectedProperty().not());

        stationAddressField.disableProperty().bind(externalProgramRadioButton.selectedProperty());
        usernameField.disableProperty().bind(externalProgramRadioButton.selectedProperty());
        passwordField.disableProperty().bind(externalProgramRadioButton.selectedProperty());
        testConnectionButton.disableProperty().bind(externalProgramRadioButton.selectedProperty());
    }

    @FXML
    private void handleAddPrefixButtonPress() {
        try {
            FXMLLoader loader = new FXMLLoader(getClass().getResource("/AddPrefixView.fxml"));
            Parent root = loader.load();

            AddPrefixController controller = loader.getController();
            controller.setStationSettingsController(this);
            controller.setDigitsPerGroup(settings.getDigitsPerGroup());

            Stage stage = new Stage();
            stage.setTitle("New prefix");
            stage.getIcons().add(new Image(getClass().getResourceAsStream("/icon.png")));
            stage.setScene(new Scene(root));
            stage.initModality(Modality.APPLICATION_MODAL);
            stage.showAndWait();
        } catch (IOException e) {
            logger.log(Level.SEVERE, "Error loading AddPrefixView", e);
        }
    }

    public void addPrefix(String prefix) {
        if (!(prefix == null || prefix.isEmpty())) {
            prefixListView.getItems().add(prefix);
        }
    }

    @FXML
    private void handleCancelButtonPress(Event e) {
        Node node = (Node) e.getSource();
        Stage stage = (Stage) node.getScene().getWindow();
        stage.close();
    }

    @FXML
    private void handleRemovePrefixButtonPress() {
        int selectedIndex = prefixListView.getSelectionModel().getSelectedIndex();
        if (selectedIndex >= 0) {
            prefixListView.getItems().remove(selectedIndex);
        }
    }

    @FXML
    private void handleSaveButtonPress(Event e) {
        if (stationName.get() == null || stationName.get().equals("")) {
            Alert alert = new Alert(Alert.AlertType.WARNING);
            alert.setTitle("Invalid station name");
            alert.setHeaderText(null);
            alert.setContentText("Station name must not be empty");
            alert.showAndWait();
            return;
        }

        if (!stationName.get().toLowerCase().equals(settings.getName().toLowerCase())) {
            for (String otherStationName : stationNames) {
                if (stationName.get().toLowerCase().equals(otherStationName.toLowerCase())) {
                    Alert alert = new Alert(Alert.AlertType.WARNING);
                    alert.setTitle("Invalid station name");
                    alert.setHeaderText(null);
                    alert.setContentText("Station name '" + stationName.get() + "' is already used.");
                    alert.showAndWait();
                    return;
                }
            }
        }

        Path oldDirectory = settings.stationPath();
        settings.setAddress(stationAddress.get());
        settings.setDigitsPerGroup(digitsPerGroup.get());
        settings.setExternalProgramCommand(externalProgramCommand.get());
        settings.setManageScheduleExternally(manageScheduleExternally.get());
        settings.setMessageLength(messageLength.get());
        settings.setMessageMethod(messageMethod.get());
        settings.setMessagePeriod(messagePeriod.get());
        settings.setName(stationName.get());
        settings.setPassword(password.get());
        settings.getPrefixes().clear();
        settings.getPrefixes().addAll(prefixListView.getItems());
        settings.setUsername(username.get());

        try {
            if (!oldDirectory.toString().equals(settings.stationPath().toString())) {
                Files.move(oldDirectory, settings.stationPath());
            }
        } catch (IOException ex) {
            logger.log(Level.SEVERE, "Failed to move directory", ex);
        }

        try {
            DateTimeFormatter formatter = DateTimeFormatter.ofPattern("HH:mm:ss");
            LocalTime startTime = LocalTime.parse(scheduleStartTimeField.getText(), formatter);
            ZonedDateTime scheduleStart = ZonedDateTime.of(scheduleStartDatePicker.getValue(), startTime, ZoneId.systemDefault());
            settings.setScheduleStart(scheduleStart.withZoneSameInstant(ZoneOffset.UTC));

        } catch (Exception ex) {
            logger.log(Level.SEVERE, "Error parsing schedule start time", ex);
        }

        String osName = System.getProperty("os.name").toLowerCase();
        if (osName.contains("win")) {
            WindowsScheduler.registerSchedule(settings);
        } else if (osName.contains("nix") || osName.contains("nux") || osName.contains("aix")) {
            LinuxScheduler.registerSchedule(settings);
        } else {
            String message = "Unsupported OS " + osName;
            logger.log(Level.SEVERE, message);

            Alert alert = new Alert(Alert.AlertType.ERROR);
            alert.setTitle(message);
            alert.setHeaderText(null);
            alert.setContentText(message);
            alert.showAndWait();
        }

        Node node = (Node) e.getSource();
        Stage stage = (Stage) node.getScene().getWindow();
        stage.setUserData(true);
        stage.close();
    }

    @FXML
    private void handleTestConnectionButtonPress() {
        if (messageMethod.get() == StationSettings.MessageMethod.SFTP) {
            String host = stationAddress.get();
            String user = username.get();
            String pass = password.get();

            connectionTestDescription.set("Testing connection");
            connectionTestStatus.set(ConnectionStatus.NEUTRAL);

            JSch jsch = new JSch();
            try {
                Session session = jsch.getSession(user, host, 22);
                session.setPassword(pass);
                session.setConfig("StrictHostKeyChecking", "no");
                session.connect(30000); // 30 seconds timeout

                if (session.isConnected()) {
                    logger.log(Level.INFO, "SFTP connection successful");
                    connectionTestDescription.set("Connection succeeded");
                    connectionTestStatus.set(ConnectionStatus.SUCCESS);
                    session.disconnect();
                } else {
                    connectionTestDescription.set("No error, but not connected");
                    connectionTestStatus.set(ConnectionStatus.FAILURE);
                    logger.log(Level.SEVERE, "SFTP connection failed");
                }
            } catch (JSchException e) {
                connectionTestDescription.set(e.getMessage());
                connectionTestStatus.set(ConnectionStatus.FAILURE);
                logger.log(Level.SEVERE, "SFTP connection failed", e);
            }
        } else if (messageMethod.get() == StationSettings.MessageMethod.WORDPRESS) {
            connectionTestDescription.set("Testing connection");
            connectionTestStatus.set(ConnectionStatus.NEUTRAL);

            System.setProperty("wordpress.username", username.get());
            System.setProperty("wordpress.password", password.get());
            System.setProperty("wordpress.url", stationAddress.get());
            Post post = new Post("Title of post", "content to post");
            try {
                WordpressClient client = new WordpressClient();
                var postTypes = client.getPostTypes();

                logger.log(Level.INFO, "Wordpress connection successful");
                connectionTestDescription.set("Connection succeeded");
                connectionTestStatus.set(ConnectionStatus.SUCCESS);
            } catch (Exception e) {
                connectionTestDescription.set("Worpress connection failed");
                connectionTestStatus.set(ConnectionStatus.FAILURE);
                logger.log(Level.SEVERE, "Wordpress connection failed", e);
            }
        }
    }

    public StringProperty stationAddressProperty() {
        return stationAddress;
    }

    public void setStationNameList(List<String> newStationNames) {
        stationNames.clear();
        stationNames.addAll(newStationNames);
    }

    public void setStationSettings(StationSettings newSettings) {
        settings = newSettings;

        stationAddress.set(settings.getAddress());
        stationName.set(settings.getName());
        digitsPerGroup.set(settings.getDigitsPerGroup());
        externalProgramCommand.set(settings.getExternalProgramCommand());
        manageScheduleExternally.set(settings.getManageScheduleExternally());
        messageLength.set(settings.getMessageLength());
        messageMethod.set(settings.getMessageMethod());
        messagePeriod.set(settings.getMessagePeriod());
        username.set(settings.getUsername());
        password.set(settings.getPassword());
        prefixListView.getItems().addAll(settings.getPrefixes());
        scheduleStartDate.set(settings.getScheduleStart().withZoneSameInstant(ZoneId.systemDefault()).toLocalDate());
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("HH:mm:ss");
        LocalTime startTime = settings.getScheduleStart().withZoneSameInstant(ZoneId.systemDefault()).toLocalTime();
        if (startTime == null) {
            startTime = LocalTime.now();
        }
        scheduleStartTime.set(startTime.format(formatter));
    }
}
